/*
 * CLOD City Lights Orchestra Device
 * 
 * v1.0 29/01/2017
 * Antoine Schmitt and CLOD Lab, Zinc/SecondeNature
 * 
 */

#define BTN_LIVE 5 // Mode LIVE
#define BTN_COMM 4 // Mode COMM
#define BTN_REC 2 // bouton rec (D4 pull-up interne)
#define THELED 15
#define LIGHTSENSOR 0

uint32_t SEUIL_MAX = 200;       // Seuil pour les écran à faible luminosité
uint32_t SEUIL_MIN = 15;       // Seuil pour les écran à faible luminosité

// mode global
#define MODE_LIVE 0
#define MODE_COMM 1
uint8_t theMode; // 0 = DIRECT, 1 = CONNECTE

// rec/play mode
#define RECORDING 0
#define PLAYING 1
#define WAITING 2
uint8_t recplayMode;
boolean recInactif;

// sequence
#define MAX_INDEX 400
#define SEQFRAMERATE 40
uint32_t deltaT_ms = 1000/SEQFRAMERATE;
uint8_t sequence[MAX_INDEX] = {0};
uint32_t nbSteps;
uint32_t sequenceIndex;
uint32_t nextStepTime;

uint32_t nextPeekTime;

#define MYDEBUG 1
uint32_t debugTime, debugDelay;
boolean debugOn = false;

// --------------------------------------------
// ----- Initialisation
// --------------------------------------------

void setup() {
  Serial.begin(115200);

  pinMode(BTN_LIVE, INPUT);
  pinMode(BTN_COMM, INPUT);
  pinMode(BTN_REC, INPUT_PULLUP);
  pinMode(THELED, OUTPUT);
  recInactif = false;

  adaptativeMinMaxInit(0, 1023);
  nextPeekTime = millis() + 1;
  
  if (MYDEBUG) {
    debugDelay = 20;
    debugTime = millis() + debugDelay;
    debugOn = false;
  }
  
  // mode à l'allumage
  theMode = modeButton();
  Serial.printf("CLOD START en mode %s\n", (theMode == MODE_LIVE ? "LIVE" : "COMM"));

  // si en mode connecté, démarrer le réseau
  if (theMode == MODE_COMM) {
    meshStart();
    jeSuisDeconnecte();
  }
}

// --------------------------------------------
// -------- La Boucle
// --------------------------------------------

void loop() {  
  if (MYDEBUG) {
    debugOn = false; // par défaut
    if (millis() >= debugTime) {
      debugOn = true; // valide sur la totalité de loop
      debugTime = millis() + debugDelay;
    }
  }
  
  
  if (theMode == MODE_LIVE) {
    // mode LIVE
    capteurLumiere();
    return;
  }
  
  if (theMode == MODE_COMM) {
    // mode COMM : en réseau Mesh
    
    // le réseau fait sa cuisine
    meshLoop();

    // pour rester adaptatif en permanence
    capteurLumiereReadOnly();
 
    // changements de mode recplayMode ?
    
    // vérifier qu'on est toujours connecté.
    if (recplayMode != WAITING && meshConnectionCount() == 0) {
      // on n'est pas en waiting, et il n'y a plus de connection = on a été déconnecté !
      jeSuisDeconnecte();
    }

    if (recplayMode == PLAYING) {
      // je suis en train de jouer
      
      if (recButtonOn() && !recInactif) {
        // on est en mode play, et le bouton REC est appuyé : on démarre l'enregistrement
        debutEnregistrement();
        
      } else if (recInactif && !recButtonOn()) {
        // on est en train de jouer, on avait forcé la fin d'enregistrement, et le bouton REC vient d'être relaché : on oublie la fin forcée
        // pour que le prochain rec fonctionne
        recInactif = false;
      }
    }

    if (recplayMode == RECORDING && !recButtonOn()) {
      // on est en mode rec et le bouton REC n'est plus appuyé : on stoppe l'enregistrement
      finEnregistrement();
    }

    // ------- LEDs, etc..
    if (recplayMode == WAITING || recplayMode == PLAYING) {
      // joue la sequence

      // est-ce l'heure de jouer la prochaine note ?
      if (millis() >= nextStepTime) {
        // y a-t-il vraiment une séquence à jouer ?
        if (nbSteps != 0 && (sequenceIndex < nbSteps)) {
        
          // récupérer la note dans la sequence, et la jouer sur la LED
          uint8_t lightVal = sequence[sequenceIndex];

          // les notes sont stockées sur un byte dans la séquence, mais la LED a une sensibilité de 4 bytes.
          uint32_t lightValReal = lightVal*4;
          analogWrite(THELED, lightValReal);
              
          // programmer la prochaine note
          sequenceIndex++;
          // boucle si nécessaire
          if (sequenceIndex >= nbSteps) sequenceIndex = 0;
        }
        
        // date de la prochaine note
        nextStepTime = millis() + deltaT_ms;
      }

    } else if (recplayMode == RECORDING) {
      // mode RECORDING
      
      // est-ce l'heure d'enregistrer la prochaine note ?
      if (millis() >= nextStepTime) {

        // récupérer la valeur lumière du capteur
        uint8_t lightVal = capteurLumiere();
        // enregistrer valeur dans la séquence
        sequence[sequenceIndex] = lightVal;
  
        // programmer la prochain enregistrement et sa date
        sequenceIndex++;

        // cas spécial de durée max
        if (sequenceIndex >= MAX_INDEX) {
          // forcer la fin de l'enregistrement : rendre le bouton Rec inactif pour que l'enregistrement ne redémarre pas immédiatement
          // Il sera inactif jusqu'à ce qu'on le relache
          finEnregistrement();
          recInactif = true;
        } else {
          // bouger le curseur de fin de sequence
          nbSteps = sequenceIndex;
          
          // prochaine date d'enregistrement
          nextStepTime = millis() + deltaT_ms;
        }
      }
    }

    return;
  }
}


// --------------------------------------------
// ------- Commandes principales
// --------------------------------------------

void jeSuisConnecte() {
  // reçu par le réseau
  Serial.printf("CLOD : connecté !! PLAYING\n");
  recplayMode = PLAYING;
    
  // reinitalise la séquence avec des pulses lents (2 seconde)
  patternTriangle(2000);
  
  rewindSequence();
}

void jeSuisDeconnecte() {
  Serial.printf("CLOD : déconnecté ;-( WAITING\n");
  recplayMode = WAITING;
  
  // reinitalise la séquence avec des pulses rapides (1/4 seconde)
  patternSquare(250);

  rewindSequence();
}

void sequenceRecue(uint8_t *seq2, int nbSteps2) {
  Serial.printf("CLOD : séquence reçue ! durée = %ds\n", nbSteps2*deltaT_ms/1000);

  // ?? que faire avec la séquence reçue ?
  // ?? que faire si je suis en train d'enregistrer ?
  if (recplayMode == RECORDING) return; // drop
  
  // exemple : oublie l'ancienne séquence, la remplace par la nouvelle
  emptySequence();
  mixSequence2(seq2, nbSteps2);
}

void envoyerSequence() {
  // seulement si quelque chose à envoyer
  if (nbSteps == 0) return;
 
  Serial.printf("CLOD : envoi séquence\n");
  meshSend(bytesToString(sequence, nbSteps));
}

void debutEnregistrement() {
  Serial.printf("CLOD : début enregistrement : RECORDING\n");
  recplayMode = RECORDING;
  rewindSequence();
}

void finEnregistrement() {
  Serial.printf("CLOD : fin d'enregistrement : PLAYING\n");
  
  recplayMode = PLAYING;
  rewindSequence();
  
  // on envoie aux autres
  envoyerSequence();
}

// ------------------------------------
// -------- les capteurs
// ------------------------------------

// Lecture du bouton MODE
uint8_t modeButton()
{
  //return MODE_COMM; // ?? TMP
  if (digitalRead(BTN_COMM)) return MODE_COMM;
  if (digitalRead(BTN_LIVE)) return MODE_LIVE;
  // par defaut : MODE_COMM
  return MODE_COMM;
}

// Lecture du bouton REC
boolean recButtonOn()
{
  boolean recOn = (digitalRead(BTN_REC) == 0);
  return recOn;
}

void capteurLumiereReadOnly() {
  // pour rester adaptatif, que la rétine continue à recevoir de la lumière tout le temps

  // n'aller voir que toutes les 100 ms
  if (millis() < nextPeekTime) return;
  nextPeekTime = millis() + 100;
  
  uint32_t sensorVal = analogRead(LIGHTSENSOR);
  uint32_t lightVal32 = adaptativeMap(sensorVal);
  
  //if (MYDEBUG && debugOn) Serial.printf("CLOD DEBUG : RAW LIGHT = %4d", sensorVal);
  //if (MYDEBUG && debugOn) Serial.printf(" (min=%3d, max=%3d) LED = %4d\n", amm_getCurMin(), amm_getCurMax(), lightVal32);
}

uint8_t capteurLumiere() {
  // avec seuils
  uint8_t lightVal = 0;
  uint32_t sensorVal = analogRead(LIGHTSENSOR);

  if (MYDEBUG && debugOn) Serial.printf("CLOD DEBUG : RAW LIGHT = %4d", sensorVal);

  // automap
  uint32_t lightVal32 = adaptativeMap(sensorVal);
  
  // car la note est sur 1 byte
  lightVal = lightVal32/4;
  
  // la LED est sur 4 bytes.
  analogWrite(THELED, lightVal32);
    
  if (MYDEBUG && debugOn) Serial.printf(" (min=%3d, max=%3d) LED = %4d", amm_getCurMin(), amm_getCurMax(), lightVal32);
  if (MYDEBUG && debugOn) Serial.printf(" note = %3d\n", lightVal);
  
  return lightVal;
}

// --------------------------------------------
// --------- Opérations sur la Sequence -------
// --------------------------------------------

void rewindSequence() {
  // rembobine la séquence
  // place le pointeur sur le dernier élément
  // programme le prochain pas à "maintenant"
  sequenceIndex = 0;
  nextStepTime = millis() - 1; // -1 juste pour être sûr que ce sera joué tout de suite
}

void emptySequence() {
  // efface la séquence
  sequenceIndex = 0;
  nbSteps = 0;
}

void mixSequence2(uint8_t *seq2, int nbSteps2) {
  // mélange une nouvelle séquence à l'ancienne
  // à l'endroit où on jouait (sequenceIndex)
  // ?? boucler si on atteint la fin de la séquence courante ? Ou ralonger ? Pour l'instant, on ralonge.
  
  for (int i = 0; i < nbSteps2 && sequenceIndex + i < MAX_INDEX; i++) {
    // pour chaque note, fait la moyenne des deux
    byte newNote = seq2[i];
    byte oldNote = sequence[sequenceIndex + i];
    int mix = (newNote + oldNote)/2;
    sequence[sequenceIndex + i] = mix;
  }
  
  // mettre à jour la longueur si elle a augmenté
  nbSteps = max(nbSteps, sequenceIndex + nbSteps2);
}

void patternSquare(int periode_ms) {
  // remplit la séquence avec un pattern LFO carré
  nbSteps = periode_ms*SEQFRAMERATE/1000;
  if (nbSteps >= MAX_INDEX) nbSteps = MAX_INDEX-1;
  for (int i = 0; i < nbSteps/2; i++) sequence[i] = 0;
  for (int i = nbSteps/2; i < nbSteps; i++) sequence[i] = 255;

}

void patternSaw(int periode_ms) {
  // remplit la séquence avec un pattern LFO Saw
  nbSteps = periode_ms*SEQFRAMERATE/1000;
  if (nbSteps >= MAX_INDEX) nbSteps = MAX_INDEX-1;
  for (int i = 0; i < nbSteps; i++) sequence[i] = i*255/nbSteps;
}

void patternTriangle(int periode_ms) {
  // remplit la séquence avec un pattern LFO Saw
  nbSteps = periode_ms*SEQFRAMERATE/1000;
  if (nbSteps >= MAX_INDEX) nbSteps = MAX_INDEX-1;
  for (int i = 0; i < nbSteps/2; i++) sequence[i] = i*255*2/nbSteps;
  for (int i = nbSteps/2; i < nbSteps; i++) sequence[i] = 2*255 - 2*255*i/nbSteps;
}


// -------------------------------------------------------
// ----- communication bas niveau avec le mesh réseau
// -------------------------------------------------------

void receivedCallback( uint32_t from, String &msg ) {
  
  //Serial.printf("mesh: Received from %d msg=%s\n", from, msg.c_str());
  uint8_t sequence2[MAX_INDEX];
  int receivedBytes;
  stringToBytes(msg, sequence2, MAX_INDEX, &receivedBytes);

  // debug
  if (MYDEBUG) {
    String printmsg = "Recu "+String(receivedBytes)+" valeurs:";
    for (int i = 0; i < receivedBytes; i++) {
      printmsg += String(sequence2[i]) + ",";
    }
    Serial.printf("mesh: Received from %d printmsg=%s\n", from, printmsg.c_str());
  }

  // traiter
  sequenceRecue(sequence2, receivedBytes);
}

void newConnectionCallback( bool adopt ) {
  int nbConn = meshConnectionCount() ;
  if (MYDEBUG) Serial.printf("mesh: New Connection, adopt=%d + nbconnections=%d\n", adopt, nbConn);
  jeSuisConnecte();
}


